﻿using System;

namespace Digiwin.Common.Torridity.Serialization {

    internal sealed class DecimalEncoder : DataEncoder<decimal> {
        /*
         * <summary>decimal存储数据的基础实现。</summary>
         * <remarks>
         * 第一个字节：
         *      0 表示后面所有的数据为0；
         *      1 表示数组都是使用1个字节存储的正整数，小于等于0xFF
         *      2 表示数组都是使用2个字节存储的正整数，小于等于0xFFFF；
         *      3 表示数组都是使用3个字节存储的正整数，小于等于0xFFFFFF；
         *      >3 表示后面的值需要一一解码；
         * 如果1-3，后续固定大小写入数据
         * 如果>3，后续字节由n个动态大小的段组成，每个段描述一个decimal：
         *      第一个字节：
         *          前5位（从左往右），正整数描述的小数位数，例如1.7位数为1，1.07位数为2，1.70位数为1；
         *          第6位，0 表示这个数就是0，无需再描述其他， 1 表示这个数是非0；
         *          第7位，0 表示是正数，1 表示是负数；
         *          第8位，0 表示使用精简描述方法描述decimal，即小数位+正负+整数部分+小数部分；
         *                  1 表示使用系统自带的方法（int[] GetBits(decimal d)）描述decimal；
         *                  注意：如果此值为1，所有前面的位无效。
         *       如果第一个字节非0，后续使用7BitEncodedInt方式编码整数部分，即使整数为0也有此字节：
         *       如果小数位> 0，还会包含7BitEncodedInt方式编码的小数部分。
         *          
         * </remarks>
         */

        #region 写入

        protected override void WriteCore(SerializeContext context, decimal[] values) {
            var firstFlag = GetFirstFlag(values);

            //首个标志位
            context.Stream.WriteByte(firstFlag);

            switch (firstFlag) {
                case 0:
                    break;
                case 1:
                    WriteToOneByte(context, values);
                    break;
                case 2:
                    WriteToTwoByte(context, values);
                    break;
                case 3:
                    WriteToThreeByte(context, values);
                    break;
                default:
                    WriteToFullByte(context, values);
                    break;
            }
        }

        private static void WriteToOneByte(SerializeContext context, decimal[] values) {
            byte[] bytes = context.GetByteCache(values.Length);
            int j = 0;
            foreach (var value in values) {
                int v = (int)value;
                bytes[j++] = (byte)v;
            }
            context.Stream.Write(bytes, 0, j);
            context.FreeBytesCache(bytes);
        }

        private static void WriteToTwoByte(SerializeContext context, decimal[] values) {
            byte[] bytes = context.GetByteCache(values.Length * 2);
            int j = 0;
            foreach (var value in values) {
                int v = (int)value;
                bytes[j++] = (byte)v;
                bytes[j++] = (byte)(v >> 0x8);
            }
            context.Stream.Write(bytes, 0, j);
            context.FreeBytesCache(bytes);
        }

        private static void WriteToThreeByte(SerializeContext context, decimal[] values) {
            byte[] bytes = context.GetByteCache(values.Length * 3);
            int j = 0;
            foreach (var value in values) {
                int v = (int)value;
                bytes[j++] = (byte)v;
                bytes[j++] = (byte)(v >> 0x8);
                bytes[j++] = (byte)(v >> 0x10);
            }
            context.Stream.Write(bytes, 0, j);
            context.FreeBytesCache(bytes);
        }

        // Fast access for 10^n where n is 0-9
        private static UInt32[] Powers10 = new UInt32[] {
            10,
            100, 
            1000, 
            10000,
            100000, 
            1000000,
            10000000,
            100000000,
            1000000000 
        };

        private static void WriteToFullByte(SerializeContext context, decimal[] values) {
            //整个逻辑放在一个方法中，目的是防止foreach循环中不断跳入其他方法，增加消耗，在这个性能要求较高的程序是允许的。
            var stream = context.Stream;
            int flag;
            int a, b, c;
            decimal d, e;

            foreach (var value in values) {
                if (value == 0m) {
                    //当数字是0时，第一个字节的所有位都是0，直接写0
                    stream.WriteByte(byte.MinValue);
                }
                else {
                    //分析出整数部分，小数位和小数部分；
                    //ex: v = 1.070m;

                    //整数部分太大，超过了int32的描述范围；
                    //当value == Int32.MinValue时，其变为正数后，比Int32.MaxValue大了1，超过范围，所以后面一个判断使用<=
                    if (value > Int32.MaxValue || value <= Int32.MinValue) {
                        goto FullWrite;
                    }

                    if (value < 0) {
                        flag = 0x4 | 0x2;
                        e = value * -1;
                        a = decimal.ToInt32(e);
                        d = e - a;
                    }
                    else {
                        flag = 0x4;
                        //取出整数部分 a = 1
                        a = decimal.ToInt32(value);
                        //得到小数部分 d = 0.070m;
                        d = value - a;
                    }

                    c = 0;
                    if (d == 0m) {
                        //没有小数部分
                        stream.WriteByte((byte)flag);
                        context.Write7BitEncodedInt(a);
                        goto NextFor;
                    }

                    do {
                        e = d * Powers10[c]; //ex: 0.7 , 7
                        b = decimal.ToInt32(e);
                        c++;

                        if (b == e) {//整数部分和小数部分相等，则得到小数部分的整数描述
                            flag = (c << 3 | flag);
                            stream.WriteByte((byte)flag);
                            context.Write7BitEncodedInt(a);
                            context.Write7BitEncodedInt(b);
                            goto NextFor;
                        }
                    } while (c < 9);

                    //小数部分太多，也使用原始方法写入。
                FullWrite:
                    stream.WriteByte((byte)1);
                    //这里使用了BinaryWriter来填充到流，此实现bw实例级缓存了byte[]并调用decimal内部的
                    //的 GetBytes 方法，这比调用GetBits不断创建数组减少了开销。
                    context.Writer.Write(value);

                NextFor:
                    flag = 0;
                }
            }
        }

        private static byte GetFirstFlag(decimal[] values) {
            decimal maxValue = 0;
            foreach (var item in values) {
                if (decimal.Floor(item) == item) {
                    if (item > maxValue) {
                        maxValue = item;
                        if (maxValue > 0xFFFFFF) {
                            return 4;//大于3个byte可描述范围 使用full的算法
                        }
                    }
                    else if (item < 0) {
                        return 4;//负数 使用string的算法
                    }
                }
                else {
                    return 4;//存在有效小数 使用full的算法
                }
            }

            if (maxValue > 0xFFFF) {
                return 3;// 65535 < maxValue <= 16777215 3个byte
            }
            else if (maxValue > 0xFF) {
                return 2;//255 < maxValue <= 65535 2个byte
            }
            else if (maxValue > 0) {//0 < maxValue <= 255 1个byte
                return 1;
            }
            return 0;
        }
        #endregion

        #region 读取

        protected override decimal[] ReadCore(SerializeContext context, int length) {
            var firstFlag = context.Stream.ReadByte();
            var values = new decimal[length];

            switch (firstFlag) {
                case 0:
                    break;
                case 1:
                    ReadFromOneByte(context, values, length);
                    break;
                case 2:
                    ReadFromTwoByte(context, values, length);
                    break;
                case 3:
                    ReadFromThreeByte(context, values, length);
                    break;
                default:
                    ReadFromFullByte(context, values, length);
                    break;
            }
            return values;
        }

        private static void ReadFromOneByte(SerializeContext context, decimal[] values, int length) {
            byte[] bytes = context.GetByteCache(length);

            context.Stream.Read(bytes, 0, length);
            for (int i = 0; i < length; i++) {
                values[i] = bytes[i];
            }
            context.FreeBytesCache(bytes);
        }

        private static void ReadFromTwoByte(SerializeContext context, decimal[] values, int length) {
            var byteCount = length * 2;
            byte[] bytes = context.GetByteCache(byteCount);
            var j = 0;

            context.Stream.Read(bytes, 0, byteCount);
            for (int i = 0; i < length; i++) {
                values[i] = (bytes[j++] | (bytes[j++] << 8));
            }
            context.FreeBytesCache(bytes);
        }

        private static void ReadFromThreeByte(SerializeContext context, decimal[] values, int length) {
            var byteCount = length * 3;
            byte[] bytes = context.GetByteCache(byteCount);
            var j = 0;

            context.Stream.Read(bytes, 0, byteCount);
            for (int i = 0; i < length; i++) {
                values[i] = ((bytes[j++] | (bytes[j++] << 8)) | (bytes[j++] << 0x10));
            }
            context.FreeBytesCache(bytes);
        }

        private static decimal[] PowersDecimal = new decimal[] {
            0.1m,
            0.01m, 
            0.001m, 
            0.0001m,
            0.00001m, 
            0.000001m,
            0.0000001m,
            0.00000001m,
            0.000000001m 
        };

        private static void ReadFromFullByte(SerializeContext context, decimal[] values, int length) {
            var stream = context.Stream;
            int flag;
            int a, b, c;
            decimal d;

            for (int i = 0; i < length; i++) {
                flag = stream.ReadByte();
                if (flag != 0) {
                    if (flag == 1) {
                        //较大的数，系统内部完整数据
                        values[i] = context.Reader.ReadDecimal();
                    }
                    else {
                        c = flag >> 3;
                        a = context.Read7BitEncodedInt();

                        if (c > 0) {
                            b = context.Read7BitEncodedInt();
                            //乘法比除法快，在测试用例中，除法时，总时间是1分04秒，而乘法时为50秒
                            d = a + ((decimal)b * PowersDecimal[c - 1]);
                        }
                        else {
                            d = a;
                        }

                        if ((flag & 0x2) == 0x2) {
                            values[i] = d * -1; //负数
                        }
                        else {
                            values[i] = d;
                        }
                    }
                }
            }
        }
        #endregion

    }
}
