﻿using System;
using System.IO;
using System.Text;
using System.Collections.Generic;

namespace Digiwin.Common.Torridity.Serialization {
    /// <summary>
    /// 序列化写入器
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1001:TypesThatOwnDisposableFieldsShouldBeDisposable")] //_writer不需要单独释放
    internal sealed class SerializeContext {
        /// <summary>
        /// 构造函数
        /// </summary>
        public SerializeContext(Stream stream)
            : this(stream, null) {
        }

        /// <summary>
        /// 构造函数
        /// </summary>
        /// <param name="encoding">用于string算法的编码器</param>
        public SerializeContext(Stream stream, Encoding encoding) {
            if (stream == null) {
                Utils.ThrowArgumentNullException("stream");
            }

            this._stream = stream;
            this._cache = new byte[0];
            this._encoding = encoding;
        }

        private readonly Stream _stream;
        public Stream Stream {
            get { return this._stream; }
        }

        private BinaryWriter _writer;
        /// <summary>
        /// 为上下文级别提供唯一实例的BinaryWriter，有助于共享内存。
        /// </summary>
        public BinaryWriter Writer {
            get {
                if (this._writer == null) {
                    this._writer = new BinaryWriter(this._stream, this.Encoding);
                }
                return this._writer;
            }
            set {
                if (value == null) {
                    Utils.ThrowArgumentNullException("value");
                }
                this._writer = value;
            }
        }

        private BinaryReader _reader;
        /// <summary>
        /// 为上下文级别提供唯一实例的BinaryReader，有助于共享内存。
        /// </summary>
        public BinaryReader Reader {
            get {
                if (this._reader == null) {
                    this._reader = new BinaryReader(this._stream, this.Encoding);
                }
                return this._reader;
            }
            set {
                if (value == null) {
                    Utils.ThrowArgumentNullException("value");
                }
                this._reader = value;
            }
        }

        private byte[] _cache;
        /// <summary>
        /// 返回一个缓存的Byte数组，用于重复使用。注意是不支持并发并小心使用。
        /// </summary>
        /// <param name="size">缓存大小</param>
        /// <returns>一个临时使用的字节数组</returns>
        public byte[] GetByteCache(int size) {
            if (_cache == null) {
                throw new InvalidOperationException();//有程序未正确调用ReleaseByteCache，或者一个程序尚未用完缓存，另外一个程序已经开始用了。
            }
            var cache = _cache;
            this._cache = null;

            if (cache.Length < size) {
                cache = new byte[size];
            }

            return cache;
        }

        /// <summary>
        /// 在用完缓存后归还缓存，仅仅是为防止程序写错了，一个缓存同时被两个程序同时使用。
        /// </summary>
        /// <param name="cache">要归还的缓存</param>
        public void FreeBytesCache(byte[] cache) {
            this._cache = cache;
        }

        private byte[] int32Bytes = new byte[5];
        /// <summary>
        /// 写入int32值,注意不支持并发
        /// </summary>
        public void Write7BitEncodedInt(int value) {
            // Write out an int 7 bits at a time.  The high bit of the byte,
            // when on, tells reader to continue reading more bytes. 
            uint v = (uint)value;   // support negative numbers
            int index = 0;

            while (v >= 0x80) {
                int32Bytes[index++] = ((byte)(v | 0x80));
                v >>= 7;
            }
            int32Bytes[index++] = ((byte)v);
            this._stream.Write(int32Bytes, 0, index);
        }

        /// <summary>
        /// 读取int值
        /// </summary>
        public int Read7BitEncodedInt() {
            // Read out an Int32 7 bits at a time.  The high bit
            // of the byte when on means to continue reading more bytes. 
            int count = 0;
            int shift = 0;
            byte b;
            do {
                // Check for a corrupted stream.  Read a max of 5 bytes.
                // In a future version, add a DataFormatException. 
                if (shift == 5 * 7)  // 5 bytes max per Int32, shift += 7 
                    throw new InvalidOperationException();

                // ReadByte handles end of stream cases for us.
                b = (byte)this._stream.ReadByte();
                count |= (b & 0x7F) << shift;
                shift += 7;
            } while ((b & 0x80) != 0);
            return count;
        }

        private Encoding _encoding;
        /// <summary>
        /// 返回当前的字符串编码器
        /// </summary>
        public Encoding Encoding {
            get {
                if (this._encoding == null) {
                    this._encoding = new UTF8Encoding(false, true);
                }
                return this._encoding;
            }
        }

        /// <summary>
        /// 编写一个字符串到流中。
        /// </summary>
        /// <param name="value">要写入的字符串，不能为null.</param>
        public void WriteString(string value) {
            if (value == null) {
                Utils.ThrowArgumentNullException("value");
            }

            int byteCount = this.Encoding.GetByteCount(value);
            byte[] bytes = this.GetByteCache(byteCount);
            this.Encoding.GetBytes(value, 0, value.Length, bytes, 0);
            this.Write7BitEncodedInt(byteCount);
            this._stream.Write(bytes, 0, byteCount);
            this.FreeBytesCache(bytes);
        }

        /// <summary>
        /// 读取一个字符串。
        /// </summary>
        /// <returns>一个字符串对象。</returns>
        public string ReadString() {
            int capacity = this.Read7BitEncodedInt();
            if (capacity < 0) {
                Utils.ThrowInvalidOperationException();
            }

            if (capacity == 0) {
                return string.Empty;
            }
            else {
                byte[] bytes = this.GetByteCache(capacity);
                this._stream.Read(bytes, 0, capacity);
                var result = this.Encoding.GetString(bytes, 0, capacity);
                this.FreeBytesCache(bytes);
                return result;
            }
        }

        private object _dictCache;
        private Type _currentDictType;
        public Dictionary<T, int> GetDictionaryCache<T>(int size) {
            //check
            if (_dictCache == null && this._currentDictType != null) {
                Utils.ThrowInvalidOperationException();//有程序未正确调用ReleaseByteCache，或者一个程序尚未用完缓存，另外一个程序已经开始用了。
            }

            Dictionary<T, int> result;
            if (typeof(T) == this._currentDictType) {

                result = (Dictionary<T, int>)_dictCache;
                this._dictCache = null;
            }
            else {
                this._dictCache = null;
                result = new Dictionary<T, int>(size);
            }

            return result;
        }

        public void FreeDictionaryCache<T>(Dictionary<T, int> dict) {
            dict.Clear();
            this._currentDictType = typeof(T);
            this._dictCache = dict;
        }

    }
}
